from keras.optimizers import Adam, RMSprop
from keras.layers import Lambda, Dense, Input
from keras.models import Model
import tensorflow as tf
from abc import abstractmethod
import keras.backend as K
import tensorflow_probability as tfp
from ..encoder import Encoder
tfd = tfp.distributions


class VariationalAutoEncoder(Encoder):
    def _sampling(self, args):
        z_mean, z_log_sigma = args
        epsilon = K.random_normal(
            shape=(self.BATCH_SIZE, self.latent_dim), mean=0.)
        return z_mean + K.exp(z_log_sigma) * epsilon

    def get_compiled_model(self, *args):
        loss_fn, lr, decay, self.observation_noise, beta = args
        input_img = Input(batch_shape=self.input_shape)

        encoder = self._encoder()
        encoded = encoder(input_img)

        # Reparametrization
        self.z_mean = Dense(self.latent_dim)(encoded)
        self.z_log_sigma = Dense(self.latent_dim)(encoded)
        z = Lambda(self._sampling, name="latent")([self.z_mean, self.z_log_sigma])

        decoder = self._decoder()
        reco = decoder(z)

        self.beta = beta
        self.loss_fn = self._get_loss(loss_fn)

        # Generator model
        decoder_input = Input(shape=(self.latent_dim,))
        _reco = decoder(decoder_input)
        generator = Model(decoder_input, _reco)

        # Model for later inference
        self.latent_model = Model(input_img, outputs=[reco, z])

        model = Model(input_img, reco)
        model.compile(optimizer=RMSprop(lr=lr, decay=decay),
                      loss=self.loss_fn, metrics=[self.KL_divergence])

        return model, generator


from abc import abstractmethod
from keras.layers import Input, Dense, PReLU, Lambda
from keras.models import Model
import keras.backend as K
from keras.optimizers import RMSprop, Adam
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions


class LadderVAE:
    def __init__(self, input_shape, latent_dim1, latent_dim2):
        self.input_shape = input_shape
        self.latent_dim1 = latent_dim1
        self.latent_dim2 = latent_dim2
        self.BATCH_SIZE = self.input_shape[0]

    def encoder1(self):
        input_tensor = Input(shape=self.input_shape[1:])
        x = Dense(512)(input_tensor)
        x = PReLU()(x)
        x = Dense(256)(x)
        x = PReLU()(x)
        x = Dense(128)(x)
        x = PReLU()(x)
        encoder = Model(input_tensor, x)
        return encoder

    def encoder2(self):
        input_tensor = Input(shape=(2 * self.latent_dim1,))
        x = Dense(256)(input_tensor)
        x = PReLU()(x)
        x = Dense(128)(x)
        x = PReLU()(x)
        x = Dense(128)(x)
        x = PReLU()(x)
        encoder = Model(input_tensor, x)
        return encoder

    def decoder2(self):
        latent2 = Input(shape=(self.latent_dim2,))
        x = Dense(256)(latent2)
        x = PReLU()(x)
        x = Dense(512)(x)
        x = PReLU()(x)
        x = Dense(1024)(x)
        x = PReLU()(x)
        reco = Dense(self.latent_dim1 * 2)(x)
        decoder = Model(latent2, reco)
        print("Decoder2")
        return decoder

    def decoder1(self):
        latent1 = Input(shape=(self.latent_dim1,))
        x = Dense(256)(latent1)
        x = PReLU()(x)
        x = Dense(512)(x)
        x = PReLU()(x)
        x = Dense(1024)(x)
        x = PReLU()(x)
        reco = Dense(self.input_shape[1])(x)
        decoder = Model(latent1, reco)
        print("Decoder1")
        return decoder

    def _split(self, _input):
        return tf.split(axis=1, value=_input, num_or_size_splits=2)

    def _sampling(self, args):
        z_mean, z_log_sigma = args
        epsilon = K.random_normal(
            shape=(self.BATCH_SIZE, self.latent_dim2 if self.mean == None else self.latent_dim1),
            mean=0 if self.mean == None else self.mean, stddev=1 if self.var == None else tf.sqrt(self.var))
        print("Epsilon : ", epsilon.get_shape())
        return z_mean + K.exp(z_log_sigma) * epsilon

    def get_compiled_model(self, *args):
        _, lr, decay, self.observation_noise, beta = args
        input_img = Input(batch_shape=self.input_shape)

        encoder1 = self.encoder1()
        encoder2 = self.encoder2()
        decoder1 = self.decoder1()
        decoder2 = self.decoder2()

        encoded1 = encoder1(input_img)

        d1 = Dense(self.latent_dim1 * 2)(encoded1)

        encoded2 = encoder2(d1)

        d2 = Dense(self.latent_dim2 * 2)(encoded2)

        # Reparametrization 1 & 2
        self.z1_mean, self.z1_log_sigma = Lambda(self._split)(d1)
        print(self.z1_mean.get_shape(), self.z1_log_sigma.get_shape())
        self.z2_mean, self.z2_log_sigma = Lambda(self._split)(d2)
        self.mean, self.var = None, None
        z2 = Lambda(self._sampling, name="latent2")(
            [self.z2_mean, self.z2_log_sigma])

        self.mean, self.var = tf.nn.moments(z2, axes=[0, 1])
        print(self.mean.get_shape(), self.var.get_shape())

        z1 = decoder2(z2)
        self.z1_mean, self.z1_log_sigma = Lambda(self._split)(z1)
        print(self.z1_mean.get_shape(), self.z1_log_sigma.get_shape())

        z1 = Lambda(self._sampling, name="latent1")(
            [self.z1_mean, self.z1_log_sigma])

        print('z2 shape : ', z2.get_shape())

        reco = decoder1(z1)

        self.beta = beta

        print('Reco done!')

        model = Model(input_img, reco)
        model.compile(optimizer=Adam(lr=lr, decay=decay),
                      loss=self.bernoulli)

        print("Compile done!")

        self.latent_model = model
        self.latent_dim = self.latent_dim2

        return model, model

    def bernoulli(self, x_true, x_reco):
        return -tf.reduce_mean(tfd.Bernoulli(x_reco)._log_prob(x_true))



from keras.optimizers import Adam
from keras.models import Model
from keras.layers import Input, Dense
from ..encoder import Encoder
from abc import abstractmethod
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions


class AutoEncoder(Encoder):
    def get_compiled_model(self, *args):
        loss_fn, lr, decay, self.observation_noise, self.beta = args
        self.beta = 0  # no KL loss (!)
        input_img = Input(batch_shape=self.input_shape)
        encoder = self._encoder()
        decoder = self._decoder()

        encoded = encoder(input_img)
        latent = Dense(self.latent_dim)(encoded)
        decoded = decoder(latent)

        # Generative model
        latent_input = Input(shape=(self.latent_dim, ))
        _decoded = decoder(latent_input)
        generative_model = Model(latent_input, _decoded)

        self.latent_model = Model(input_img, outputs=[decoded, latent])

        self.loss_fn = self._get_loss(loss_fn)

        model = Model(input_img, decoded)
        model.compile(optimizer=Adam(lr=lr, decay=decay), loss=self.loss_fn)
        return model, generative_model



from keras.preprocessing.image import ImageDataGenerator
import numpy as np
from .data_loader import DataLoader


class DataGenerator:
    def __init__(self, image_shape, batch_size, file_path='/data/scramtex_700_28px.pkl', binarize=False):
        self.DATA_LOADER = DataLoader(file_path, binarize)
        self.IMAGE_SHAPE = image_shape
        self.BATCH_SIZE = batch_size
        self.DATA_GENERATOR = ImageDataGenerator()
        self.TRAIN, self.TEST = self.DATA_LOADER.train_test_split()

    def flow(self):
        return self.DATA_GENERATOR.flow(self.TRAIN, self.TRAIN, batch_size=self.BATCH_SIZE)

    def contrast_flow(self):
        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                yield (self._contrast(batch_x), batch_y)
        return train_generator(self.flow())

    def flattened_flow(self):
        image_dim = np.prod(self.IMAGE_SHAPE)

        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                yield batch_x.reshape(self.BATCH_SIZE, image_dim), batch_y.reshape(self.BATCH_SIZE, image_dim)
        return train_generator(self.flow())

    def flattened_contrast_flow(self):
        image_dim = np.prod(self.IMAGE_SHAPE)

        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                batch_x, batch_y = self._contrast(
                    batch_x), self._contrast(batch_y)
                yield batch_x.reshape(self.BATCH_SIZE, image_dim), batch_y.reshape(self.BATCH_SIZE, image_dim)

        return train_generator(self.flow())

    def _contrast(self, images):
        contrasted_images = np.zeros(shape=images.shape)
        for ind in range(images.shape[0]):

            contrasted_images[ind] = np.clip(
                np.random.rand() * 2. * (images[ind] - 0.5) + 0.5, 0, 1)
        return contrasted_images

    def validation_data(self):
        return self.TEST, self.TEST

    def flattened_validation_data(self):
        image_dim = np.prod(self.IMAGE_SHAPE)
        reshaped_test = self.TEST.reshape(self.TEST.shape[0], image_dim)
        return reshaped_test, reshaped_test



import pickle
from sklearn import model_selection
import numpy as np
import os
from sklearn.preprocessing import StandardScaler


class DataLoader:
    def __init__(self, file_path, binarize):
        self.binarize = binarize
        with open(file_path, 'rb') as f:
            self.data = pickle.load(f)

    def train_test_split(self):
        try:
            X_train = self.data['train_images']
            X_test = self.data['test_images']

            X_train = np.clip(X_train, 0, 1.0)
            X_test = np.clip(X_test, 0, 1.0)

            mean, std = np.mean(X_train), np.std(X_train)
            print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
            print('Min: %.3f, Max: %.3f' % (np.min(X_train), np.max(X_train)))
            print('Training size : %d \t Test size : %d' %
                  (X_train.shape[0], X_test.shape[0]))

            channel = int(np.prod(X_train.shape) /
                          (X_train.shape[0] * 28 * 28))

            X_train, _ = model_selection.train_test_split(
                X_train.reshape(X_train.shape[0], 28, 28, channel), test_size=0, random_state=45)
            X_test, _ = model_selection.train_test_split(
                X_test.reshape(X_test.shape[0], 28, 28, channel), test_size=0, random_state=42)
        except IndexError:
            X = self.data
            X = np.clip(X, 0, 1.0)

            mean, std = np.mean(X), np.std(X)
            print('Mean: %.3f, Standard Deviation: %.3f' % (mean, std))
            print('Min: %.3f, Max: %.3f' % (np.min(X), np.max(X)))

            channel = int(np.prod(X.shape) / (X.shape[0] * 28 * 28))

            X_train, X_test = model_selection.train_test_split(
                X.reshape(X.shape[0], 28, 28, channel), test_size=0.05, random_state=137)
        finally:
            if self.binarize:
                X_train[X_train >= .5] = 1.
                X_train[X_train < .5] = 0.
                X_test[X_test >= .5] = 1.
                X_test[X_test < .5] = 0.
            return (X_train, X_test)
